/*
 * Copyright (c) 2003, 2004
 *	Damien Bergamini <damien.bergamini@free.fr>. All rights reserved.
 * Copyright (c) 2005 Matthieu Castet <castet.matthieu@free.fr>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice unmodified, this list of conditions, and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 */
#include <linux/init.h>
#include <linux/crc32.h>
#include <linux/usb.h>
#include <linux/firmware.h>
#include <linux/kthread.h>

#include "debug.h"

#include "ueaglevar.h"
#include "ueaglereg.h"

#define UPDATE_ATM_STAT(type, val) if (sc->usbatm->atm_dev) \
					sc->usbatm->atm_dev->type = val;

//static char *dsp[NB_MODEM] = {[0 ... (NB_MODEM - 1)] = "DSP.bin" };
static char *cmv[NB_MODEM] = {[0 ... (NB_MODEM - 1)] = NULL };

//module_param_array(dsp, charp, NULL, 0644);
//MODULE_PARM_DESC(dsp, "dsp file");
module_param_array(cmv, charp, NULL, 0644);
MODULE_PARM_DESC(cmv, "cmv file");

/*
 * Make sure that the DSP code provided by eaglectl is safe to use.
 */
static int check_dsp(u8 *dsp, int len)
{
	u8 *p, *end, *pp;
	u8 pagecount, blockcount;
	u16 blocksize;
	u32 pageoffset;
	int i, j;

	end = dsp + len;

	/* enough space for pagecount? */
	if (dsp + 1 > end)
		return 1;

	p = dsp;
	pagecount = FW_GET_BYTE(p);
	p += 1;

	/* enough space for page offsets? */
	if (p + 4 * pagecount > end)
		return 1;

	for (i = 0; i < pagecount; i++) {

		pageoffset = FW_GET_LONG(p);
		p += 4;

		if (pageoffset == 0)
			continue;

		/* enough space for blockcount? */
		if (dsp + pageoffset + 1 > end)
			return 1;

		pp = dsp + pageoffset;
		blockcount = FW_GET_BYTE(pp);
		pp += 1;

		for (j = 0; j < blockcount; j++) {

			/* enough space for block header? */
			if (pp + 4 > end)
				return 1;

			pp += 2;	/* skip blockaddr */
			blocksize = FW_GET_WORD(pp);
			pp += 2;

			/* enough space for block data? */
			if (pp + blocksize > end)
				return 1;

			pp += blocksize;
		}
	}

	return 0;
}

/* 
 * send data to the idma pipe 
 * */
static int uea_idma_write(struct uea_softc *sc, void *data, u32 size)
{
	int ret = -ENOMEM;
	u8 *xfer_buff;
	int bytes_read;

	xfer_buff = kmalloc(size, GFP_KERNEL);
	if (!xfer_buff) {
		uea_err(INS_TO_USBDEV(sc), "can't allocate xfer_buff\n");
		return ret;
	}

	memcpy(xfer_buff, data, size);

	ret =
	    usb_bulk_msg(sc->usb_dev,
			 usb_sndbulkpipe(sc->usb_dev, UEA_IDMA_PIPE),
			 xfer_buff, size, &bytes_read, BULK_TIMEOUT);

	kfree(xfer_buff);
	if (ret < 0) {
		return ret;
	}
	if (size != bytes_read) {
		uea_err(INS_TO_USBDEV(sc), "size != bytes_read %d %d\n", size,
		       bytes_read);
		return -EIO;
	}

	return 0;
}

static int request_dsp(struct uea_softc *sc)
{
	int ret;
	char *dsp_name;

	if (UEA_CHIP_VERSION(sc) == ADI930) {
		if (IS_ISDN(sc))
			dsp_name = "DSP9i.bin";
		else
			dsp_name = "DSP9p.bin";
	}
	else {
		if (IS_ISDN(sc))
			dsp_name = "DSPei.bin";
		else
			dsp_name = "DSPep.bin";
	}
	ret = request_firmware(&sc->dsp_firm, dsp_name, &sc->usb_dev->dev);
	if (ret < 0) {
		uea_err(INS_TO_USBDEV(sc),
		       "requesting firmware %s failed with error %d\n",
		       dsp_name, ret);
		return ret;
	}

	if (check_dsp(sc->dsp_firm->data, sc->dsp_firm->size)) {
		uea_err(INS_TO_USBDEV(sc), "firmware %s is corrupted\n",
		       dsp_name);
		release_firmware(sc->dsp_firm);
		sc->dsp_firm = NULL;
		return -EILSEQ;
	}

	return 0;
}

/*
 * The uea_load_page() function must be called within a process context
 */
static void uea_load_page(void *xsc)
{
	struct uea_softc *sc = xsc;
	u16 pageno = sc->pageno;
	u16 ovl = sc->ovl;
	block_info_t bi;

	u8 *p;
	u8 pagecount, blockcount;
	u16 blockaddr, blocksize;
	u32 pageoffset;
	int i;

	if (sc->dsp_firm == NULL && request_dsp(sc) < 0)
		return;

	p = sc->dsp_firm->data;
	pagecount = FW_GET_BYTE(p);
	p += 1;

	if (pageno >= pagecount)
		goto bad1;

	p += 4 * pageno;
	pageoffset = FW_GET_LONG(p);

	if (pageoffset == 0)
		goto bad1;

	p = sc->dsp_firm->data + pageoffset;
	blockcount = FW_GET_BYTE(p);
	p += 1;

	uea_dbg(DBG_BOOT, INS_TO_USBDEV(sc),
	       "sending %u blocks for DSP page %u\n", blockcount, pageno);

	bi.wHdr = cpu_to_le16(UEA_BIHDR);
	bi.wOvl = cpu_to_le16(ovl);
	bi.wOvlOffset = cpu_to_le16(ovl | 0x8000);

	for (i = 0; i < blockcount; i++) {
		blockaddr = FW_GET_WORD(p);
		p += 2;

		blocksize = FW_GET_WORD(p);
		p += 2;

		bi.wSize = cpu_to_le16(blocksize);
		bi.wAddress = cpu_to_le16(blockaddr);
		bi.wLast = cpu_to_le16((i == blockcount - 1) ? 1 : 0);

		/* send block info through the IDMA pipe */
		if (uea_idma_write(sc, &bi, BLOCK_INFO_SIZE))
			goto bad2;

		/* send block data through the IDMA pipe */
		if (uea_idma_write(sc, p, blocksize))
			goto bad2;

		p += blocksize;
	}

	return;

      bad2:
	uea_err(INS_TO_USBDEV(sc), "sending DSP block %u failed\n", i);
	return;
      bad1:
	uea_err(INS_TO_USBDEV(sc), "invalid DSP page %u requested\n",pageno);
}

#define UEA_READ_CMV(sc, address, offset, _data) \
	do { \
		int ret; \
		ret = uea_cmv(sc, MAKEFUNCTION(MEMACCESS, REQUESTREAD), \
				address, offset, 0); \
		if (ret < 0) { \
			uea_err(INS_TO_USBDEV(sc), \
				"reading cmv failed with error %d\n", ret); \
			return ret; \
		} \
		_data = sc->data; \
	} while(0)

#define UEA_WRITE_CMV(sc, address, offset, data) \
	do { \
		int ret; \
		ret = uea_cmv(sc, MAKEFUNCTION(MEMACCESS, REQUESTWRITE), \
				address, offset, data); \
		if (ret < 0) { \
			uea_err(INS_TO_USBDEV(sc), \
				"writing cmv failed with error %d\n", ret); \
			return ret; \
		} \
	} while(0)

#define WAKE_UP_CMV_ACK() \
	do { \
		sc->cmv_ack = 1; \
		wake_up_interruptible(&sc->cmv_ack_wait); \
	} while (0)

static inline int wait_cmv_ack(struct uea_softc *sc)
{
	int ret = wait_event_interruptible_timeout(sc->cmv_ack_wait,
						   sc->cmv_ack, ACK_TIMEOUT);
	sc->cmv_ack = 0;

	if (ret < 0)
		return ret;

	return (ret == 0) ? -ETIMEDOUT : 0;

}

#define UCDC_SEND_ENCAPSULATED_COMMAND 0x00

int uea_request(struct uea_softc *sc, 
		u16 value, u16 index, u16 size, void *data)
{
	u8 *xfer_buff;
	int ret = -ENOMEM;

	xfer_buff = kmalloc(size, GFP_KERNEL);
	if (!xfer_buff) {
		uea_err(INS_TO_USBDEV(sc), "can't allocate xfer_buff\n");
		return ret;
	}
	memcpy(xfer_buff, data, size);

	ret = usb_control_msg(sc->usb_dev, usb_sndctrlpipe(sc->usb_dev, 0),
			      UCDC_SEND_ENCAPSULATED_COMMAND,
			      USB_DIR_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			      value, index, xfer_buff, size, CTRL_TIMEOUT);

	kfree(xfer_buff);
	if (ret < 0) {
		uea_err(INS_TO_USBDEV(sc), "usb_control_msg error %d\n", ret);
		return ret;
	}

	if (ret != size) {
		uea_err(INS_TO_USBDEV(sc),
		       "usb_control_msg send only %d bytes (instead of %d)\n",
		       ret, size);
		return -EIO;
	}

	return 0;
}

static int uea_cmv(struct uea_softc *sc, 
		u8 function, u32 address, u16 offset, u32 data)
{
	cmv_t cmv;
	int ret;

	sc->cmv_idx++;
	sc->cmv_address = address;
	sc->cmv_offset = offset;

	cmv.wPreamble = cpu_to_le16(PREAMBLE);
	cmv.bDirection = HOSTTOMODEM;
	cmv.bFunction = function;
	cmv.wIndex = cpu_to_le16(sc->cmv_idx);
	put_unaligned(cpu_to_le32(address), &cmv.dwSymbolicAddress);
	cmv.wOffsetAddress = cpu_to_le16(offset);
	put_unaligned(cpu_to_le32(data >> 16 | data << 16), &cmv.dwData);

	ret = uea_request(sc, UEA_SET_BLOCK, UEA_MPTX_START, CMV_SIZE, &cmv);
	if (ret < 0)
		return ret;
	return wait_cmv_ack(sc);
}

/* 
 * Monitor the modem and update the stat
 * return 0 if everything is ok
 * return < 0 if an error occurs (-EAGAIN reboot needed)
 */
static int uea_stat(struct uea_softc *sc)
{
	u32 data;

	data = sc->stats.phy.state;
	UEA_READ_CMV(sc, SA_STAT, 0, sc->stats.phy.state);

	switch ((sc->stats.phy.state >> 8) & 0xf) {
	case 0:		/* not yet synchronized */
		uea_dbg(DBG_STAT, INS_TO_USBDEV(sc),
		       "modem not yet synchronized\n");
		return 0;

	case 1:		/* initialization */
		uea_dbg(DBG_STAT, INS_TO_USBDEV(sc), "modem initializing\n");
		return 0;

	case 2:		/* operational */
		uea_dbg(DBG_STAT, INS_TO_USBDEV(sc), "modem operational\n");
		break;

	case 3:		/* fail ... */
		uea_info(INS_TO_USBDEV(sc), "modem synchronization failed\n");
		return -EAGAIN;

	case 4 ... 6:	/* test state */
		uea_warn(INS_TO_USBDEV(sc), 
				"modem in test mode - not supported\n");
		return -EAGAIN;

	case 7:		/* fast-retain ... */
		uea_info(INS_TO_USBDEV(sc), "modem in fast-retain mode\n");
		return 0;
	default:
		uea_err(INS_TO_USBDEV(sc), "modem invalid SW mode %d\n",
		       (sc->data >> 8) & 0xf);
		return -EAGAIN;
	}
	if ((((data >> 8) & 0xf) != 2)) {
		uea_request(sc, UEA_SET_MODE, UEA_LOOPBACK_OFF, 0, NULL);
		uea_info(INS_TO_USBDEV(sc), "modem operational\n");

		/* release the dsp firmware as it is not needed until
		 * the next failure
		 */
		if (sc->dsp_firm) {
			release_firmware(sc->dsp_firm);
			sc->dsp_firm = NULL;
		}

		UEA_READ_CMV(sc, SA_INFO, 10, data);
		uea_info(INS_TO_USBDEV(sc), "ATU-R firmware version : %x\n",
				data);
	}
	/* always update it as atm layer could not be init when we switch to
	 * operational state
	 */
	UPDATE_ATM_STAT(signal, ATM_PHY_SIG_FOUND);
	/* wake up processes waiting for synchronization */
	wake_up_interruptible(&sc->sync_q);

	UEA_READ_CMV(sc, SA_DIAG, 2, sc->stats.phy.flags);
	sc->stats.phy.mflags |= sc->stats.phy.flags;

#if 0
	if (sc->data & 0x10) {
		DPRINTF(("delineation LOSS\n"));
		return -EAGAIN;
	}
#else
	/* in case of a flags (delineation LOSS (& 0x10)), we check the status
	 * again in order to detect the failure earlier
	 */
	if (sc->stats.phy.flags) {
		uea_dbg(DBG_STAT, INS_TO_USBDEV(sc), "Stat flag = %d\n",
		       sc->stats.phy.flags);
		return 0;
	}
#endif

	UEA_READ_CMV(sc, SA_RATE, 0, data);
	/* in bulk mode the modem have problem with high rate
	 * changing internal timing could improve things, but the 
	 * value is misterious
	 */
	if (sc->usbatm->driver->in == UEA_BULK_DATA_PIPE
		    && sc->stats.phy.dsrate != (data >> 16) * 32) {
		/* Original timming from ADI(used in windows driver)
		 * 0x20ffff>>16 * 32 = 32 * 32 = 1Mbits
		 */
		int timeout = (data <= 0x20ffff) ? 0 : 1;
		uea_request(sc, UEA_SET_TIMEOUT, timeout, 0, NULL);
		uea_info(INS_TO_USBDEV(sc), 
				"setting new timeout %d\n", timeout);
	}
	sc->stats.phy.dsrate = (data >> 16) * 32;
	sc->stats.phy.usrate = (data & 0xffff) * 32;
	UPDATE_ATM_STAT(link_rate, sc->stats.phy.dsrate * 1000 / 424);

	UEA_READ_CMV(sc, SA_DIAG, 3, sc->stats.phy.cells_crc_errors);

	UEA_READ_CMV(sc, SA_DIAG, 22, sc->stats.phy.dserror);

	UEA_READ_CMV(sc, SA_DIAG, 23, data);
	sc->stats.phy.attenuation = (data & 0xff) / 2;

	UEA_READ_CMV(sc, SA_DIAG, 25, sc->stats.phy.dsmargin);

	UEA_READ_CMV(sc, SA_DIAG, 46, sc->stats.phy.userror);

	UEA_READ_CMV(sc, SA_DIAG, 49, sc->stats.phy.usmargin);

	UEA_READ_CMV(sc, SA_DIAG, 51, sc->stats.phy.rxflow);

	UEA_READ_CMV(sc, SA_DIAG, 52, sc->stats.phy.txflow);

	UEA_READ_CMV(sc, SA_DIAG, 54, sc->stats.phy.dsunc);
	/* DIAG53 ack */
	/* near_end_Current_Corrected_Blocks */

	UEA_READ_CMV(sc, SA_DIAG, 58, sc->stats.phy.usunc);

	UEA_READ_CMV(sc, SA_INFO, 8, sc->stats.phy.vidco);

	UEA_READ_CMV(sc, SA_INFO, 14, sc->stats.phy.vidcpe);

	return 0;
}

int request_cmvs(struct uea_softc *sc,
		 struct uea_cmvs **cmvs, const struct firmware **fw)
{
	int ret, size;
	u8 *data;
	char *cmv_name;

	if (cmv[sc->dev] == NULL) {
		if (UEA_CHIP_VERSION(sc) == ADI930) {
			if (IS_ISDN(sc))
				cmv_name = "CMV9i.bin";
			else
				cmv_name = "CMV9p.bin";
		}
		else {
			if (IS_ISDN(sc))
				cmv_name = "CMVei.bin";
			else
				cmv_name = "CMVep.bin";
		}
	}
	else
		cmv_name = cmv[sc->dev];

	ret = request_firmware(fw, cmv_name, &sc->usb_dev->dev);
	if (ret < 0) {
		uea_err(INS_TO_USBDEV(sc),
		       "requesting firmware %s failed with error %d\n",
		       cmv_name, ret);
		return ret;
	}

	data = (u8 *) (*fw)->data;
	size = *data * sizeof(struct uea_cmvs) + 1;
	if (size != (*fw)->size) {
		uea_err(INS_TO_USBDEV(sc), "firmware %s is corrupted\n",
		       cmv_name);
		release_firmware(*fw);
		return -EILSEQ;
	}

	*cmvs = (struct uea_cmvs *)(data + 1);
	return *data;
}

/*-
 * 1) tell the modem that we want to boot in IDMA mode
 * 2) enter reset mode
 * 3) wait 100ms
 * 4) leave reset mode
 * 5) clear tx and rx mailboxes
 * 6) wait 1s before boot straping the firmware
 * 7) wait for modem ready CMV
 * 8) send CMV options
 */

static int uea_start_reset(struct uea_softc *sc)
{
	u16 zero = 0;	/* ;-) */
	int i, len;
	struct uea_cmvs *cmvs;
	const struct firmware *fw;

	uea_enters(DBG_BOOT, INS_TO_USBDEV(sc));
	uea_info(INS_TO_USBDEV(sc), "(re)booting started\n");

	sc->booting = 1;
	UPDATE_ATM_STAT(signal, ATM_PHY_SIG_LOST);
	/* reset statistics */
	memset(&sc->stats, 0, sizeof(struct uea_stats));

	uea_request(sc, UEA_SET_MODE, UEA_LOOPBACK_ON, 0, NULL);
	uea_request(sc, UEA_SET_MODE, UEA_BOOT_IDMA, 0, NULL);
	uea_request(sc, UEA_SET_MODE, UEA_START_RESET, 0, NULL);

	/* original driver use 200ms, but windows driver use 100ms */
	msleep(100);

	uea_request(sc, UEA_SET_MODE, UEA_END_RESET, 0, NULL);
	uea_request(sc, UEA_SET_2183_DATA, UEA_MPTX_MAILBOX, 2, &zero);
	uea_request(sc, UEA_SET_2183_DATA, UEA_MPRX_MAILBOX, 2, &zero);
	uea_request(sc, UEA_SET_2183_DATA, UEA_SWAP_MAILBOX, 2, &zero);

	msleep(1000);
	sc->cmv_idx = 0;
	sc->booting = 0;

	len = request_cmvs(sc, &cmvs, &fw);

	/* reload firmware if the firmware is still loaded */
	if (sc->dsp_firm) {
		release_firmware(sc->dsp_firm);
		sc->dsp_firm = NULL;
	}
	sc->pageno = 0;
	sc->ovl = 0;
	schedule_work(&sc->task);

	/* wait for modem ready CMV */
	wait_cmv_ack(sc);

	/* send options */
	for (i = 0; i < len; i++)
		UEA_WRITE_CMV(sc, FW_GET_LONG(&cmvs[i].address),
			      FW_GET_WORD(&cmvs[i].offset),
			      FW_GET_LONG(&cmvs[i].data));

	if (len >= 0)
		release_firmware(fw);

	sc->reset = 0;
	uea_leaves(DBG_BOOT, INS_TO_USBDEV(sc));
	return 0;
}

/*
 * In case of an error wait 1s before rebooting the modem
 * if the modem don't request reboot (-EAGAIN).
 * Monitor the modem every 1s
 */
static int uea_kthread(void *data)
{
	struct uea_softc *sc = (struct uea_softc *)data;
	int ret = -EAGAIN;
	uea_enters(DBG_STAT, INS_TO_USBDEV(sc));
	while (!kthread_should_stop()) {
		if (ret < 0 || sc->reset)
			ret = uea_start_reset(sc);
		if (!ret)
			ret = uea_stat(sc);
		if (ret != -EAGAIN)
			msleep_interruptible(1000);
	}
	uea_leaves(DBG_STAT, INS_TO_USBDEV(sc));
	return ret;
}

/* Load second usb firmware for ADI930 chip */
static int load_XILINX_firmware(struct uea_softc *sc)
{
	const struct firmware *fw_entry;
	int ret, size, u, ln;
	u8 *pfw;
	char *fw_name = "930-fpga.bin";
#if 0
	u32 crc = 0;
#endif

	uea_enters(DBG_INIT, INS_TO_USBDEV(sc));

	ret = request_firmware(&fw_entry, fw_name, &sc->usb_dev->dev);
	if (ret) {
		uea_err(INS_TO_USBDEV(sc), "firmware %s is not available\n",
		       fw_name);
		goto err0;
	}

	pfw = fw_entry->data;
	size = fw_entry->size;
#if 0
	crc = FW_GET_LONG(pfw);
	pfw += 4;
	size -= 4;

	if (crc32_be(0, pfw, size) != crc) {
#else
	if (size != 0x577B) {
#endif
		uea_err(INS_TO_USBDEV(sc), "firmware %s is corrupted\n",
		       fw_name);
		ret = -EILSEQ;
		goto err1;
	}
	for (u = 0; u < size; u += ln) {
		ln = min(size - u, 64);
		ret = uea_request(sc, 0xe, 0, ln, pfw + u);
		if (ret < 0) {
			uea_err(INS_TO_USBDEV(sc),
			       "elsa download data failed (%d)\n", ret);
			goto err1;
		}
	}

	ret = uea_request(sc, 0xe, 1, 0, NULL);
	if (ret < 0)
		uea_err(INS_TO_USBDEV(sc), "elsa de-reset failure (%d)\n", ret);

      err1:
	release_firmware(fw_entry);
      err0:
	uea_leaves(DBG_INIT, INS_TO_USBDEV(sc));
	return ret;
}

static void uea_dispatch_cmv(struct uea_softc *sc, cmv_t * cmv)
{
	uea_enters(DBG_INTS, INS_TO_USBDEV(sc));
	if (le16_to_cpu(cmv->wPreamble) != PREAMBLE)
		goto bad1;

	if (cmv->bDirection != MODEMTOHOST)
		goto bad1;

	if (cmv->bFunction == MAKEFUNCTION(ADSLDIRECTIVE, MODEMREADY) &&
	    !sc->cmv_idx) {
		WAKE_UP_CMV_ACK();
		return;
	}

	if (le16_to_cpu(cmv->wIndex) != sc->cmv_idx ||
	    le32_to_cpu(get_unaligned(&cmv->dwSymbolicAddress)) !=
	    sc->cmv_address
	    || le16_to_cpu(cmv->wOffsetAddress) != sc->cmv_offset)
		goto bad2;

	if (FUNCTION_TYPE(cmv->bFunction) != MEMACCESS)
		goto bad2;

	if (FUNCTION_SUBTYPE(cmv->bFunction) != REPLYREAD &&
	    FUNCTION_SUBTYPE(cmv->bFunction) != REPLYWRITE)
		goto bad2;

	sc->data = le32_to_cpu(get_unaligned(&cmv->dwData));
	sc->data = sc->data << 16 | sc->data >> 16;

	WAKE_UP_CMV_ACK();
	return;

      bad2:
	uea_err(INS_TO_USBDEV(sc), "unexpected cmv received\n");
	return;

      bad1:
	uea_err(INS_TO_USBDEV(sc), "invalid cmv received, "
			"wPreamble %d, bDirection %d\n", 
			le16_to_cpu(cmv->wPreamble), cmv->bDirection);
}

/* 
 * interrupt handler
 */
static void uea_intr(struct urb *urb, struct pt_regs *regs)
{
	struct uea_softc *sc = (struct uea_softc *)urb->context;
	intr_pkt_t *intr;
	uea_enters(DBG_INTS, INS_TO_USBDEV(sc));

	if (urb->status < 0) {
		uea_err(INS_TO_USBDEV(sc), "uea_intr() failed with %d\n",
		       urb->status);
		return;
	}

	intr = (intr_pkt_t *) urb->transfer_buffer;

	/* device-to-host interrupt */
	if (intr->bType != 0x08 || sc->booting) {
		uea_err(INS_TO_USBDEV(sc), "wrong intr\n");
		// rebooting ?
		// sc->reset = 1;
		goto resubmit;
	}

	switch (le16_to_cpu(intr->wInterrupt)) {
	case INT_LOADSWAPPAGE:
		sc->pageno = intr->bSwapPageNo;
		sc->ovl = intr->bOvl >> 4 | intr->bOvl << 4;
		schedule_work(&sc->task);
		break;

	case INT_INCOMINGCMV:
		uea_dispatch_cmv(sc, &intr->u.s2.cmv);
		break;

	default:
		uea_err(INS_TO_USBDEV(sc), "unknown intr %u\n",
		       le16_to_cpu(intr->wInterrupt));
	}

      resubmit:
	usb_submit_urb(sc->urb_int, GFP_ATOMIC);
}

/*
 * Start the modem : init the data and start kernel thread
 */
int uea_boot(struct uea_softc *sc)
{
	int ret;
	intr_pkt_t *intr;

	uea_enters(DBG_BOOT, INS_TO_USBDEV(sc));

	INIT_WORK(&sc->task, uea_load_page, sc);
	init_waitqueue_head(&sc->sync_q);
	init_waitqueue_head(&sc->cmv_ack_wait);

	if (UEA_CHIP_VERSION(sc) == ADI930)
		load_XILINX_firmware(sc);

	intr = kmalloc(INTR_PKT_SIZE, GFP_KERNEL);
	if (!intr) {
		uea_err(INS_TO_USBDEV(sc),
		       "cannot allocate interrupt package\n");
		uea_leaves(DBG_BOOT, INS_TO_USBDEV(sc));
		return -ENOMEM;
	}

	sc->urb_int = usb_alloc_urb(0, GFP_KERNEL);
	if (!sc->urb_int) {
		uea_err(INS_TO_USBDEV(sc), "cannot allocate interrupt URB\n");
		goto err;
	}

	usb_fill_int_urb(sc->urb_int, sc->usb_dev,
			 usb_rcvintpipe(sc->usb_dev, UEA_INTR_PIPE),
			 intr, INTR_PKT_SIZE, uea_intr, sc,
			 sc->usb_dev->actconfig->interface[0]->altsetting[0].
			 endpoint[0].desc.bInterval);

	ret = usb_submit_urb(sc->urb_int, GFP_KERNEL);
	if (ret < 0) {
		uea_err(INS_TO_USBDEV(sc),
		       "urb submition failed with error %d\n", ret);
		goto err1;
	}

	sc->kthread = kthread_run(uea_kthread, sc, "ueagle-atm");
	if (sc->kthread == ERR_PTR(-ENOMEM)) {
		uea_err(INS_TO_USBDEV(sc), "failed to create thread\n");
		goto err2;
	}

	uea_leaves(DBG_BOOT, INS_TO_USBDEV(sc));
	return 0;

      err2:
	usb_kill_urb(sc->urb_int);
      err1:
	kfree(intr);
      err:
	usb_free_urb(sc->urb_int);
	uea_leaves(DBG_BOOT, INS_TO_USBDEV(sc));
	return -ENOMEM;
}

/*
 * Stop the modem : kill kernel thread and free data
 */
void uea_stop(struct uea_softc *sc)
{
	int ret;
	uea_enters(DBG_INIT, INS_TO_USBDEV(sc));
	ret = kthread_stop(sc->kthread);
	uea_info(INS_TO_USBDEV(sc), "kthread finish with status %d\n", ret);

	/* stop any pending boot process */
	flush_scheduled_work();

     	uea_request(sc, UEA_SET_MODE, UEA_LOOPBACK_ON, 0, NULL);

	usb_kill_urb(sc->urb_int);
	kfree(sc->urb_int->transfer_buffer);
	usb_free_urb(sc->urb_int);

	if (sc->dsp_firm)
		release_firmware(sc->dsp_firm);
	uea_leaves(DBG_INIT, INS_TO_USBDEV(sc));
}
