/*-
 * Copyright (c) 2005 Matthieu Castet <castet.matthieu@free.fr>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#define EAGLEUSBVERSION "ueagle-svn $Id: ueagle.c 63 2005-06-26 20:20:06Z matc $"

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/crc32.h>
#include <linux/usb.h>
#include <linux/firmware.h>
#include <linux/ctype.h>

#include "debug.h"

#include "ueaglevar.h"
#include "ueaglereg.h"

static int uea_read_proc(char *page, char **start, off_t off, int count,
			int *eof, void *data);

static struct usb_driver uea_driver;
static struct proc_dir_entry *uea_procdir;
DECLARE_MUTEX(uea_semaphore);

/*
 * User supplied debug level
 */
unsigned int module_dbg_mask = 0x0;

static int dev = 0;
static int iso[NB_MODEM] = {[0 ... (NB_MODEM - 1)] = 1 };
static int wait[NB_MODEM] = {[0 ... (NB_MODEM - 1)] = 0 };

module_param(module_dbg_mask, uint, 0644);
MODULE_PARM_DESC(module_dbg_mask, "Module Debug mask");
module_param_array(iso, bool, NULL, 0644);
MODULE_PARM_DESC(iso, "iso mode");
module_param_array(wait, bool, NULL, 0644);
MODULE_PARM_DESC(wait, "wait the synchronisation before starting ATM");

/* Firmware loading */
#define LOAD_INTERNAL     0xA0
#define F8051_USBCS       0x7f92
/**
 * uea_send_modem_cmd - Send a command to the modem (firmware loading)
 *
 */
static int uea_send_modem_cmd(struct usb_device *usb, 
		u16 addr, u16 count, u8 * buff)
{
	int ret = -ENOMEM;
	u8 *xfer_buff;

	xfer_buff = kmalloc(count, GFP_KERNEL);
	if (xfer_buff) {
		memcpy(xfer_buff, buff, count);
		ret = usb_control_msg(usb,
				      usb_sndctrlpipe(usb, 0),
				      LOAD_INTERNAL,
				      USB_DIR_OUT | USB_TYPE_VENDOR |
				      USB_RECIP_DEVICE, addr, 0, xfer_buff,
				      count, CTRL_TIMEOUT);
		kfree(xfer_buff);
	}

	if (ret < 0)
		return ret;

	return (ret == count) ? 0 : -EIO;
}

/**
 * uea_load_firmware - Load firmware into pre-firmware devices.
 *
 */
static int uea_load_firmware(struct usb_device *usb, unsigned int ver)
{
	int ret, size;
	const struct firmware *fw_entry;
	u8 *pfw, value;
	char *fw_name = "eagle.fw";
	u32 crc = 0;

	uea_enters(DBG_INIT, usb);
	uea_info(usb, "pre-firmware device, uploading firmware\n");

	switch (ver) {
	case ADI930:
		fw_name = "adi930.fw";
		break;
	case EAGLE_I:
		fw_name = "eagleI.fw";
		break;
	case EAGLE_II:
		fw_name = "eagleII.fw";
		break;
	case EAGLE_III:
		fw_name = "eagleIII.fw";
		break;
	}

	ret = request_firmware(&fw_entry, fw_name, &usb->dev);
	if (ret) {
		/* sometime hotplug don't have time to give the firmware the
		 * first time. Retry it.
		 */
		uea_err(usb, "firmware %s is not available, retring...\n", 
				fw_name);
		msleep(1000);
		ret = request_firmware(&fw_entry, fw_name, &usb->dev);
		if (ret) {
			uea_err(usb, "firmware %s is not available\n", fw_name);
			goto err0;
		}
	}
	pfw = fw_entry->data;
	size = fw_entry->size;

	crc = FW_GET_LONG(pfw);
	pfw += 4;
	size -= 4;
	if (crc32_be(0, pfw, size) != crc) {
		uea_err(usb, "firmware %s is corrupted\n", fw_name);
		ret = -EILSEQ;
		goto err1;
	}

	/*
	 * Send reset
	 */
	value = 1;
	ret = uea_send_modem_cmd(usb, F8051_USBCS, sizeof(value), &value);

	if (ret < 0) {
		uea_err(usb, "modem reset failed with error %d\n", ret);
		goto err1;
	}

	while (size > 0) {
		u8 len = FW_GET_BYTE(pfw);
		u16 add = FW_GET_WORD(pfw + 1);
		ret = uea_send_modem_cmd(usb, add, len, pfw + 3);
		if (ret < 0) {
			uea_err(usb, "uploading firmware data failed " 
					"with error %d\n", ret);
			goto err1;
		}
		pfw += len + 3;
		size -= len + 3;
	}

	/*
	 * De-assert reset
	 */
	value = 0;
	ret = uea_send_modem_cmd(usb, F8051_USBCS, 1, &value);
	if (ret < 0)
		uea_err(usb, "modem de-reset failed with error %d\n", ret);
	else
		uea_info(usb, "firmware uploaded\n");

      err1:
	release_firmware(fw_entry);
      err0:
	uea_leaves(DBG_INIT, usb);
	return ret;
}

/* syfs interface */
static struct uea_softc *dev_to_uea(struct device *dev)
{
	struct usb_interface *intf;
	struct usbatm_data *usbatm;

	intf = to_usb_interface(dev);
	if (!intf)
		return NULL;

	usbatm = usb_get_intfdata(intf);
	if (!usbatm)
		return NULL;

	return (struct uea_softc *)usbatm->driver_data;
}

#define UEA_SYSFS_OP_ENTER \
	struct uea_softc *sc; \
	down(&uea_semaphore); \
	sc = dev_to_uea(dev); \
	if (!sc) { \
		up(&uea_semaphore); \
		return -ENODEV; \
	} \

#define UEA_SYSFS_OP_LEAVE \
	up(&uea_semaphore);

static ssize_t read_status(struct device *dev, char *buf)
{
	u32 data;
	UEA_SYSFS_OP_ENTER data = sc->stats.phy.state;

	UEA_SYSFS_OP_LEAVE return snprintf(buf, 9, "%8.8x\n", data);
}

static ssize_t reboot(struct device *dev, const char *buf, size_t count)
{
	UEA_SYSFS_OP_ENTER sc->reset = 1;

	UEA_SYSFS_OP_LEAVE return count;
}

static DEVICE_ATTR(status, S_IWUGO | S_IRUGO, read_status, reboot);

#define UEA_ATTR(name, reset) \
 \
static ssize_t read_##name(struct device *dev, char *buf) \
{ \
	u32 data; \
	UEA_SYSFS_OP_ENTER \
 \
	data = sc->stats.phy.name; \
	if (reset) \
		sc->stats.phy.name = 0; \
 \
	UEA_SYSFS_OP_LEAVE \
	return snprintf(buf, 9, "%8.8x\n", data);\
} \
\
static DEVICE_ATTR(name, S_IRUGO, read_##name, NULL)

UEA_ATTR(mflags, 1);

/* Retrieve the device End System Identifier (MAC) */

#define htoi(x) (isdigit(x) ? x-'0' : toupper(x)-'A'+10)
static int uea_getesi(struct uea_softc *sc, u_char * esi)
{
	unsigned char mac_str[2 * ETH_ALEN + 1];
	int i;
	if (usb_string
	    (sc->usb_dev, sc->usb_dev->descriptor.iSerialNumber, mac_str,
	     sizeof(mac_str)) != 2 * ETH_ALEN)
		return 1;

	for (i = 0; i < ETH_ALEN; i++)
		esi[i] = htoi(mac_str[2 * i]) * 16 + htoi(mac_str[2 * i + 1]);

	return 0;
}

/* ATM stuff */
static int uea_atm_open(struct usbatm_data *usbatm, struct atm_dev *atm_dev)
{
	struct uea_softc *sc = (struct uea_softc *)usbatm->driver_data;

	return uea_getesi(sc, atm_dev->esi);
}

static void uea_atm_stop(struct usbatm_data *usbatm, struct atm_dev *atm_dev)
{
	//struct uea_softc *sc = (struct uea_softc *)usbatm->driver_data;
}

static int uea_heavy(struct usbatm_data *usbatm, struct usb_interface *intf)
{
	struct uea_softc *sc = (struct uea_softc *)usbatm->driver_data;
	if (wait_event_interruptible(sc->sync_q,
				     ((sc->stats.phy.state >> 8) & 0xf) == 2))
		return -EINTR;

	return 0;

}

static int claim_interface(struct usb_device *usb_dev,
			   struct usbatm_data *usbatm, int ifnum)
{
	int ret;
	struct usb_interface *intf = usb_ifnum_to_if(usb_dev, ifnum);

	if (!intf) {
		uea_err(usb_dev, "interface %d not found\n", ifnum);
		return -ENODEV;
	}

	ret = usb_driver_claim_interface(&uea_driver, intf, usbatm);
	if (ret != 0)
		uea_err(usb_dev, "can't claim interface %d, error %d\n", ifnum,
		       ret);
	return ret;
}

static void create_fs_entries(struct uea_softc *sc, struct usb_interface *intf)
{
	char path[32];
	struct proc_dir_entry *pde;

	/* procfs interface */
	if (uea_procdir) {
		int busnum = sc->usb_dev->bus->busnum;
		int devnum = sc->usb_dev->devnum;

		snprintf(path, 32, "%03d-%03d", busnum, devnum);
		pde = create_proc_read_entry(path,
					     0, uea_procdir, uea_read_proc, sc);

		if (pde) {
			pde->owner = THIS_MODULE;
			uea_info(INS_TO_USBDEV(sc), 
					"created proc entry at: "
					"/proc/driver/ueagle-atm/%s\n", path);
		} else
			uea_err(INS_TO_USBDEV(sc), 
					"can't create proc entry at: "
					"/proc/driver/ueagle-atm/%s\n", path);
	}

	/* sysfs interface */
	device_create_file(&intf->dev, &dev_attr_status);
	device_create_file(&intf->dev, &dev_attr_mflags);
}

static int uea_bind(struct usbatm_data *usbatm, struct usb_interface *intf,
		   const struct usb_device_id *id, int *heavy)
{
	struct usb_device *usb = interface_to_usbdev(intf);
	struct uea_softc *sc;
	int ret, ifnum = intf->altsetting->desc.bInterfaceNumber;

	uea_enters(DBG_INIT, usb);

	if (ifnum != 0)
		return -ENODEV;

	*heavy = wait[dev];

	/* interface 1 is for outbound traffic */
	ret = claim_interface(usb, usbatm, 1);
	if (ret < 0)
		return ret;

	/* interface 2 is for inbound traffic */
	ret = claim_interface(usb, usbatm, 2);
	if (ret < 0)
		return ret;

	sc = kcalloc(1, sizeof(struct uea_softc), GFP_KERNEL);
	if (!sc) {
		uea_err(INS_TO_USBDEV(sc), "uea_init: not enough memory !\n");
		return -ENOMEM;
	}

	sc->usb_dev = usb;
	usbatm->driver_data = sc;
	sc->usbatm = usbatm;
	sc->dev = (dev < NB_MODEM)?dev++:0;
	sc->driver_info = id->driver_info;

	ret = uea_boot(sc);
	if (ret < 0) {
		kfree(sc);
		return ret;
	}

	create_fs_entries(sc, intf);
	return 0;
}

static void destroy_fs_entries(struct uea_softc *sc, struct usb_interface *intf)
{
	char path[32];

	/* procfs interface */
	if (uea_procdir) {
		int busnum = sc->usb_dev->bus->busnum;
		int devnum = sc->usb_dev->devnum;

		snprintf(path, 32, "%03d-%03d", busnum, devnum);
		remove_proc_entry(path, uea_procdir);
	}

	/* sysfs interface */
	device_remove_file(&intf->dev, &dev_attr_status);
	device_remove_file(&intf->dev, &dev_attr_mflags);
}

static void uea_unbind(struct usbatm_data *usbatm, struct usb_interface *intf)
{
	struct uea_softc *sc = (struct uea_softc *)usbatm->driver_data;

	destroy_fs_entries(sc, intf);
	uea_stop(sc);
	kfree(sc);
}

static struct usbatm_driver euatm_drv = {
	.driver_name = "ueagle-atm",
	.owner = THIS_MODULE,
	.bind = uea_bind,
	.atm_start = uea_atm_open,
	.unbind = uea_unbind,
	.atm_stop = uea_atm_stop,
	.heavy_init = uea_heavy,
	.in = UEA_BULK_DATA_PIPE,
	.out = UEA_BULK_DATA_PIPE,
};

static int uea_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
	struct usb_device *usb = interface_to_usbdev(intf);

	uea_enters(DBG_INIT, usb);
	uea_info(usb, "ADSL device founded vid (%#X) pid (%#X) \n",
	       le16_to_cpu(usb->descriptor.idVendor),
	       le16_to_cpu(usb->descriptor.idProduct));

	usb_reset_device(usb);

	if (UEA_IS_PREFIRM(id))
		return uea_load_firmware(usb, UEA_CHIP_VERSION(id));

	if (iso[dev < NB_MODEM?dev:0]) {
		//XXX try other alternate insterface it failed
		/*
		 * Set alternate interface to 8, which is ISO transport
		 * with the max. packet size ( about 1007 bytes)
		 */
		if (usb_set_interface(usb, USB_INTF_IN, FASTEST_ISO_INTF) < 0) {
			uea_err(usb, "usb_set_interface failed on iso alt 8\n");
		} else {
			euatm_drv.in = UEA_ISO_DATA_PIPE;
			euatm_drv.use_iso = 1;
			uea_info(usb, "iso mode\n");
		}
	}

	return usbatm_usb_probe(intf, id, &euatm_drv);
}

static void uea_disconnect(struct usb_interface *intf)
{
	struct usb_device *usb = interface_to_usbdev(intf);
	int ifnum = intf->altsetting->desc.bInterfaceNumber;
	uea_enters(DBG_INIT, usb);

	if (usb->config->desc.bNumInterfaces == 3 && ifnum == 0) {
		down(&uea_semaphore);
		usbatm_usb_disconnect(intf);
		up(&uea_semaphore);
		uea_info(usb, "ADSL device removed\n");
	}

	uea_leaves(DBG_INIT, usb);
}

/**
 * uea_read_proc  - /proc interface.
 *
 */
static int uea_read_proc(char *page, 
		char **start, off_t off, int count, int *eof, void *data)
{
	struct uea_softc *sc = (struct uea_softc *)data;
	char *p = page;

	/*
	 * If this output needs to be larger than 4K (PAGE_SIZE), 
	 * we need to do this differently
	 */

	*start = NULL;
	*eof = 1;
	if (off != 0)
		return 0;

	p += sprintf(p, "eagle-usb status display\n");
	p += sprintf(p,
		     "-------------------------------------------------------------\n");
	p += sprintf(p, "Driver version %s     Chipset: Eagle%1.1d\n",
		     EAGLEUSBVERSION, UEA_CHIP_VERSION(sc));

	if (sc->usb_dev != NULL) {
		p += sprintf(p,
			     "Vendor ID : 0x%x     Product ID : 0x%x   Rev: 0x%x(%s)\n",
			     le16_to_cpu(sc->usb_dev->descriptor.idVendor),
			     le16_to_cpu(sc->usb_dev->descriptor.idProduct),
			     le16_to_cpu(sc->usb_dev->descriptor.bcdDevice),
			     IS_ISDN(sc)? "isdn" : "pots");
		p += sprintf(p,
			     "USB Bus : %03d\t USB Device : %03d\t Dbg mask: 0x%x\n",
			     sc->usb_dev->bus->busnum, sc->usb_dev->devnum,
			     module_dbg_mask);
	}

	p += sprintf(p, "Tx Rate   %10d  Rx Rate   %10d\n",
		     sc->stats.phy.usrate, sc->stats.phy.dsrate);
	p += sprintf(p, "Tx Margin %10d  Rx Margin %10d   Atten   %10d dB\n",
		     sc->stats.phy.usmargin,
		     sc->stats.phy.dsmargin, sc->stats.phy.attenuation);
	p += sprintf(p, "Tx error  %10d  Rx error  %10d\n",
		     sc->stats.phy.userror, sc->stats.phy.dserror);
	p += sprintf(p, "Tx FEC    %10d  Rx FEC    %10d\n",
		     sc->stats.phy.usunc, sc->stats.phy.dsunc);
	p += sprintf(p, "Tx flow   %10d  Rx flow   %10d\n",
		     sc->stats.phy.txflow, sc->stats.phy.rxflow);
	p += sprintf(p, "VID-CPE   %10d  VID-CO    %10d    HEC    %10d\n",
		     sc->stats.phy.vidcpe,
		     sc->stats.phy.vidco, sc->stats.phy.cells_crc_errors);

	p += sprintf(p, "Delin          ");

	/*
	 * Delineation is the only one where we print a string 
	 * instead of a number 
	 */

	if (sc->stats.phy.flags & 0x0C00)
		p += sprintf(p, "ERROR");
	else if (sc->stats.phy.flags & 0x0030)
		p += sprintf(p, " LOSS");
	else
		p += sprintf(p, " GOOD");

	p += sprintf(p, "  Flags     %10.10x\n", sc->stats.phy.flags);
	switch ((sc->stats.phy.state >> 8) & 0xf) {
	case 1:
		p += sprintf(p, "Modem is initializing\n");
		break;
	case 0:
		p += sprintf(p, "Modem is booting\n");
		break;
	case 2:
		p += sprintf(p, "Modem is operational\n");
		break;
	default:
		p += sprintf(p, "Modem synchronization failed\n");
		break;
	}
	p += sprintf(p, "\n");

	return p - page;
}

#define UEA_DEVICE(vid, pid, info) \
	{USB_DEVICE((vid), (pid)), .driver_info = (info)}

/*
 * List of supported VID/PID
 */
static const struct usb_device_id uea_ids[] = {
	UEA_DEVICE(ELSA_VID,	ELSA_PID_PREFIRM,	ADI930 | PREFIRM),
	UEA_DEVICE(ELSA_VID,	ELSA_PID_PSTFIRM,	ADI930 | PSTFIRM),
	UEA_DEVICE(EAGLE_VID,	EAGLE_I_PID_PREFIRM,	EAGLE_I | PREFIRM),
	UEA_DEVICE(EAGLE_VID,	EAGLE_I_PID_PSTFIRM,	EAGLE_I | PSTFIRM),
	UEA_DEVICE(EAGLE_VID,	EAGLE_II_PID_PREFIRM,	EAGLE_II | PREFIRM),
	UEA_DEVICE(EAGLE_VID,	EAGLE_II_PID_PSTFIRM,	EAGLE_II | PSTFIRM),
	UEA_DEVICE(EAGLE_VID,	EAGLE_IIC_PID_PREFIRM,	EAGLE_II | PREFIRM),
	UEA_DEVICE(EAGLE_VID,	EAGLE_IIC_PID_PSTFIRM,	EAGLE_II | PSTFIRM),
	UEA_DEVICE(EAGLE_VID,	EAGLE_III_PID_PREFIRM,	EAGLE_III | PREFIRM),
	UEA_DEVICE(EAGLE_VID,	EAGLE_III_PID_PSTFIRM,	EAGLE_III | PSTFIRM),
	UEA_DEVICE(USR_VID,	MILLER_A_PID_PREFIRM,	EAGLE_I | PREFIRM),
	UEA_DEVICE(USR_VID,	MILLER_A_PID_PSTFIRM,	EAGLE_I | PSTFIRM),
	UEA_DEVICE(USR_VID,	MILLER_B_PID_PREFIRM,	EAGLE_I | PREFIRM),
	UEA_DEVICE(USR_VID,	MILLER_B_PID_PSTFIRM,	EAGLE_I | PSTFIRM),
	UEA_DEVICE(USR_VID,	HEINEKEN_A_PID_PREFIRM, EAGLE_I | PREFIRM),
	UEA_DEVICE(USR_VID,	HEINEKEN_A_PID_PSTFIRM, EAGLE_I | PSTFIRM),
	UEA_DEVICE(USR_VID,	HEINEKEN_B_PID_PREFIRM,	EAGLE_I | PREFIRM),
	UEA_DEVICE(USR_VID,	HEINEKEN_B_PID_PSTFIRM,	EAGLE_I | PSTFIRM),
	{}
};

/*
 * USB driver descriptor
 */
static struct usb_driver uea_driver = {
	.owner = THIS_MODULE,
	.name = "ueagle-atm",
	.id_table = uea_ids,
	.probe = uea_probe,
	.disconnect = uea_disconnect,
};

MODULE_DEVICE_TABLE(usb, uea_ids);

/**
 * uea_init - Initialize the module.
 *      Creates /proc/driver/eagle-usb directory
 *      Register to USB subsystem
 *
 */
static int __init uea_init(void)
{
	printk(KERN_INFO "[ueagle-usb] driver " EAGLEUSBVERSION " loaded\n");

	uea_procdir = proc_mkdir("driver/ueagle-atm", NULL);
	if (uea_procdir)
		uea_procdir->owner = THIS_MODULE;
	else
		printk(KERN_ERR
		       "[UEAGLE-USB]  could not create /proc/driver/eagle-usb/\n");

	usb_register(&uea_driver);

	return 0;
}

module_init(uea_init);

/**
 * uea_exit  -  Finalize module
 *
 *    Deregister with USB subsystem
 *    Remove /proc/drivers/eagle-usb directory
 *
 */
static void __exit uea_exit(void)
{
	/*
	 * This calls automatically the uea_disconnect method if necessary:
	 */
	usb_deregister(&uea_driver);

	printk(KERN_INFO "[ueagle-usb] driver unloaded\n");

	/* remove non existing proc entries is ok */
	remove_proc_entry("driver/ueagle-atm", NULL);
}

module_exit(uea_exit);

MODULE_AUTHOR("Frederick Ros/Matthieu Castet/Damien Bergamini");
MODULE_DESCRIPTION("Eagle USB ADSL Modem driver");
MODULE_LICENSE("GPL");
